# Copyright (c) 2010 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, string, errno
from subprocess import Popen, PIPE, STDOUT
from functools import wraps

from PyQt4 import QtCore
from pythoncom import CoInitialize, CoUninitialize
from wmi import WMI
from fdisk.device import Device
from fdisk.mbr import MBRDisk

from rapidwriter.backends.base import BaseDisk, BaseBackend
from rapidwriter.util import (
  QThreadWithEventLoop,
  FunctionObject,
  locate_data_file,
)
from rapidwriter.exceptions import (
  DeviceError,
  DeviceNotFoundError,
  DeviceBusyError,
  SubprocessError,
)
from rapidwriter.logger import log_info, log_error, log_traceback


class WindowsDisk(BaseDisk):
    wmi_object_path = None

    def __init__(self, wmi_object_path):
        super(WindowsDisk, self).__init__()
        self.wmi_object_path = wmi_object_path

    ###

    def get_name(self):
        return u'PhysicalDrive%u' % self.get_wmi_disk_drive().Index

    def get_manufacturer(self):
        return self.get_wmi_disk_drive().Manufacturer

    def get_model(self):
        return self.get_wmi_disk_drive().Model

    def get_size(self):
        return long(self.get_wmi_disk_drive().Size)

    def is_removable(self):
        return (7 in self.get_wmi_disk_drive().capabilities)

    def get_device_path(self):
        return self.get_wmi_disk_drive().DeviceID

    ###

    def get_wmi_disk_drive(self):
        return WMI().get(self.wmi_object_path)

    def get_wmi_partition(self):
        return self.get_wmi_disk_drive().associators(
          wmi_result_class = 'Win32_DiskPartition',
        )[0]

    def get_wmi_logical_disk(self):
        return self.get_wmi_partition().associators(
          wmi_result_class = 'Win32_LogicalDisk',
        )[0]

    def get_volume(self):
        # FIXME: Volume name? (\\?\Volume{UUID})
        device_id = self.get_wmi_logical_disk().DeviceID
        drive_letter = device_id[0]
        if drive_letter not in string.uppercase:
            raise AssertionError(device_id)
        return '%s:' % drive_letter

    def _exec(self, args):
        popen = Popen(args, stdout = PIPE, stderr = STDOUT, shell = True)
        try:
            popen.wait()
            output = popen.stdout.read()
        finally:
            popen.stdout.close()
        if popen.returncode != 0:
            log_error(output)
            raise SubprocessError(popen.returncode, output)

    ###

    def create_partition_table(self):
        try:
            device = Device(self.get_device_path())
        except IOError, e:
            log_traceback()

            msg = unicode(str(e), errors = 'replace')
            if e.errno == errno.ENOENT:
                raise DeviceNotFoundError(msg)
            raise DeviceError(msg)

        try:
            sector_count = device.get_sector_count()

            # Force 255 heads, 63 sectors no matter what the current partition
            # table looks like.  I'm not sure if this is a good idea or not.
            # We'll see.
            geometry = MBRDisk.calculate_geometry(sector_count, 255, 63)
            mbr_disk = MBRDisk(geometry)

            with open(locate_data_file('mbr.bin', ('arch', 'pkg')), 'rb') as f:
                mbr_disk.boot_record.code.read_from_file(f)

            mbr_disk.add_partition(type = 0x0b)

            try:
                mbr_disk.write_to_device(device)
            except IOError, e:
                log_traceback()

                msg = unicode(str(e), errors = 'replace')
                if e.errno == errno.EBUSY:
                    raise DeviceBusyError(msg)
                raise DeviceError(msg)

        finally:
            device.close()

    def create_filesystem(self):
        # FIXME: Maybe we can do this via WMI?
        # Note: Must use shell on win32 to perform PATH searching.
        # See CreateProcess documentation for details.
        self._exec(['FORMAT', self.get_volume(), '/FS:FAT32', '/X', '/Y', '/Q'])

    def install_bootloader(self):
        syslinux = locate_data_file('syslinux.exe', ('arch', 'pkg'))
        self._exec([syslinux, self.get_volume()])

    def mount(self):
        mount_point = os.path.join(self.get_volume(), '\\')
        # All we need to do is access the volume and it will be mounted for us.
        os.listdir(mount_point)
        return mount_point

    def umount(self, mount_point):
        # This could be done via a Windows DeviceIoControl call.
        pass

    def sync(self):
        # A lock/unlock cycle will cause the filesystem to be dismounted, which
        # should flush any buffers.  It'd be nice to simplify this with, e.g.,
        # a single dismount call, but this will do for now.  (I'm not sure if
        # we even need to worry about it on Windows, actually.)
        device = Device(self.get_device_path())
        device.lock()
        device.unlock()


class WindowsDeviceManager(QtCore.QObject):
    backend = None
    disks = None

    def __init__(self, backend):
        super(WindowsDeviceManager, self).__init__()
        self.backend = backend
        self.disks = []

    @QtCore.pyqtSlot()
    def run(self):
        # FIXME: Is this thread terminated safely?
        CoInitialize()
        try:
            wmi_conn = WMI(find_classes = False)
            watcher = wmi_conn.Win32_DiskDrive.watch_for()
            for wmi_disk_drive in wmi_conn.Win32_DiskDrive():
                self._add_wmi_disk_drive(wmi_disk_drive)
            while True:
                event = watcher()
                if event.event_type == 'creation':
                    self._maybe_add_wmi_disk_drive(event)
                elif event.event_type == 'deletion':
                    self._maybe_remove_wmi_disk_drive(event)
        finally:
            CoUninitialize()

    def _maybe_add_wmi_disk_drive(self, wmi_disk_drive):
        log_info('wmi: device created: %r', wmi_disk_drive.DeviceID)
        try:
            self._find_disk(wmi_disk_drive)
        except DeviceNotFoundError:
            self._add_wmi_disk_drive(wmi_disk_drive)

    def _add_wmi_disk_drive(self, wmi_disk_drive):
        log_info('wmi: adding device: %r', wmi_disk_drive.DeviceID)
        disk = WindowsDisk(str(wmi_disk_drive.path()))
        self.disks.append(disk)
        FunctionObject.call_in_main_thread(
          lambda: self.backend.add_disk(disk)
        )

    def _maybe_remove_wmi_disk_drive(self, wmi_disk_drive):
        log_info('wmi: device deleted: %r', wmi_disk_drive.DeviceID)
        try:
            disk = self._find_disk(wmi_disk_drive)
        except DeviceNotFoundError:
            pass
        else:
            self.disks.remove(disk)
            # FIXME: We should probably invalidate the disk object somehow,
            # right?  I guess things will fail when the next WMI lookup is
            # attempted.
            FunctionObject.call_in_main_thread(
              lambda: self.backend.remove_disk(disk)
            )

    def _find_disk(self, wmi_disk_drive):
        wmi_object_path = str(wmi_disk_drive.path())
        for disk in self.disks:
            if disk.wmi_object_path == wmi_object_path:
                return disk
        raise DeviceNotFoundError()


class WindowsBackend(BaseBackend):
    _thread = None
    _mgr = None

    def __init__(self, *args, **kwargs):
        super(WindowsBackend, self).__init__(*args, **kwargs)

        self._thread = QThreadWithEventLoop()

        self._mgr = WindowsDeviceManager(self)
        self._mgr.moveToThread(self._thread)

        self._thread.started.connect(self._mgr.run)

    def start(self):
        self._thread.start()

    def process_thread(self, thread):
        thread.run = self._wrap_thread_run(thread.run)

    def _wrap_thread_run(self, fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            CoInitialize()
            try:
                return fn(*args, **kwargs)
            finally:
                CoUninitialize()
        return wrapped
